#version 430

// Input attributes

in vec4 v2fColor;
in vec3 v2fNormal;
in vec3 v2fDiffuseSpecularEmissive;
in vec2 v2fTexCoord;
in vec3 v2fPos;
in vec3 v2fTexIndex;

// Uniform data

struct Light 
{
    vec3 pos;
    vec3 color;
    float strength;
};

const int maxLights = 128;

layout( location = 2 ) uniform vec3 uSceneAmbient;
layout( location = 3 ) uniform vec3 uViewDir;
layout( location = 4 ) uniform int lightCount;
layout( location = 5 ) uniform Light lights[maxLights];

layout( binding = 0 ) uniform sampler2D uTexture0;
layout( binding = 1 ) uniform sampler2D uTexture1;
layout( binding = 2 ) uniform sampler2D uTexture2;
layout( binding = 3 ) uniform sampler2D uTexture3;
layout( binding = 4 ) uniform sampler2D uTexture4;
layout( binding = 5 ) uniform sampler2D uTexture5;
layout( binding = 6 ) uniform sampler2D uTexture6;
layout( binding = 7 ) uniform sampler2D uTexture7;

// Fragment shader outputs

layout( location = 0 ) out vec4 oColor;

// In-shader vars
vec4 diffuseColor = vec4(1, 1, 1, 1);
vec4 specularColor = vec4(1, 1, 1, 1);
vec4 emissiveColor = vec4(1, 1, 1, 1);

// Function definitions
void choose_texture();


void main() 
{
    // Determine texture components
    choose_texture();    

    // Determine ambient component of lighting
    vec4 colorLinear = vec4(uSceneAmbient, 1) * diffuseColor;


    // Apply each point light's compenents
    for (int i = 0; i < lightCount; i++)
    {
        // Unpack properties
        vec3 lightPos = lights[i].pos;
        vec4 lightColor = vec4(lights[i].color, 1);
        float lightStrength = lights[i].strength;

        // Determine properties for lighting calculations
        vec3 normal = normalize(v2fNormal);
        vec3 lightDir = lightPos - v2fPos;
        float lightDistance = length(lightDir);

        lightDistance = lightDistance * lightDistance; 
        lightDir = normalize(lightDir);

        float lDotN = max(dot(lightDir, normal), 0.0); // Important to ensure > 0

        vec3 viewDir = normalize(uViewDir);
        vec3 halfDir = normalize(lightDir + viewDir);

        float specAngle = max(dot(halfDir, normal), 0.0);  // Important to ensure > 0
        float specular = pow(specAngle, v2fDiffuseSpecularEmissive.y);

        // Combine calculated properties into Blinn-Phong lighting
        colorLinear += (diffuseColor * lightColor * lDotN * lightStrength * v2fDiffuseSpecularEmissive.x) / lightDistance + // Diffuse component
                       (specularColor * lightColor * specular * lightStrength * v2fDiffuseSpecularEmissive.y) / lightDistance + // Specular Component
                       emissiveColor * v2fDiffuseSpecularEmissive.z; // Emmissive Component
    }

    // Combine color with material color and texture color
    colorLinear = colorLinear * v2fColor;
	
	// Output full RGBA color value
	oColor = vec4(colorLinear.x, colorLinear.y, colorLinear.z, colorLinear.w);
}

void choose_texture()
{
    // There is probably a better way to do this, but it works (for now)
    // Can't define an array of sampler2Ds and access by index,
    // and don't want to call texture() on each one then choose for performance reasons
    if (v2fTexIndex.x == 0.0)
	{
		diffuseColor = texture( uTexture0, v2fTexCoord ).rgba;
	}
    else if (v2fTexIndex.x == 1.0)
    {
        diffuseColor = texture( uTexture1, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.x == 2.0)
    {
        diffuseColor = texture( uTexture2, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.x == 3.0)
    {
        diffuseColor = texture( uTexture3, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.x == 4.0)
    {
        diffuseColor = texture( uTexture4, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.x == 5.0)
    {
        diffuseColor = texture( uTexture5, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.x == 6.0)
    {
        diffuseColor = texture( uTexture6, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.x == 7.0)
    {
        diffuseColor = texture( uTexture7, v2fTexCoord ).rgba;
    }
    
    if (v2fTexIndex.y == 0.0)
	{
		specularColor = texture( uTexture0, v2fTexCoord ).rgba;
	}
    else if (v2fTexIndex.y == 1.0)
    {
        specularColor = texture( uTexture1, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.y == 2.0)
    {
        specularColor = texture( uTexture2, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.y == 3.0)
    {
        specularColor = texture( uTexture3, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.y == 4.0)
    {
        specularColor = texture( uTexture4, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.y == 5.0)
    {
        specularColor = texture( uTexture5, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.y == 6.0)
    {
        specularColor = texture( uTexture6, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.y == 7.0)
    {
        specularColor = texture( uTexture7, v2fTexCoord ).rgba;
    }

    if (v2fTexIndex.z == 0.0)
	{
		emissiveColor = texture( uTexture0, v2fTexCoord ).rgba;
	}
    else if (v2fTexIndex.z == 1.0)
    {
        emissiveColor = texture( uTexture1, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.z == 2.0)
    {
        emissiveColor = texture( uTexture2, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.z == 3.0)
    {
        emissiveColor = texture( uTexture3, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.z == 4.0)
    {
        emissiveColor = texture( uTexture4, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.z == 5.0)
    {
        emissiveColor = texture( uTexture5, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.z == 6.0)
    {
        emissiveColor = texture( uTexture6, v2fTexCoord ).rgba;
    }
    else if (v2fTexIndex.z == 7.0)
    {
        emissiveColor = texture( uTexture7, v2fTexCoord ).rgba;
    }
}